//***************************************************************************
//
// CONIO32.CPP: Template for 32-bit console programs
//
// Written by:	Daniel D. Miller
//		Jason Hood
//
// Last update: 11/07/01 14:44
//		12 Jan, 2015
//
//***************************************************************************
// Windows console data structures
//
// typedef struct _CONSOLE_SCREEN_BUFFER_INFO { // csbi
// COORD      dwSize;
// COORD      dwCursorPosition;
// WORD       wAttributes;
// SMALL_RECT srWindow;
// COORD      dwMaximumWindowSize;
// } CONSOLE_SCREEN_BUFFER_INFO ;
//
// typedef struct _SMALL_RECT { // srct
// SHORT Left;
// SHORT Top;
// SHORT Right;
// SHORT Bottom;
// } SMALL_RECT;
//
// typedef struct _COORD { // coord.
// SHORT X;      // horizontal coordinate
// SHORT Y;      // vertical coordinate
// } COORD;
//***************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>	// _getch()
#include "ndata.hpp"
#include "conio32.hpp"
#include <wchar.h>


HANDLE hStdOut;
int    redirected;

static CONSOLE_SCREEN_BUFFER_INFO sinfo;
#define curpos sinfo.dwCursorPosition

static DWORD original_mode;
static WORD  original_attribs;
static bool  wrapped;

// from NDATA.CPP
extern int reportcols;

//***************************************************************************
//***************************************************************************
// GENERIC 32-BIT CONSOLE I/O FUNCTIONS
//***************************************************************************
//***************************************************************************

//**********************************************************
BOOL control_handler( DWORD dwCtrlType )
{
  restore_console_attribs();
  return FALSE;
}

//***************************************************************************
// This stores CONSOLE_SCREEN_BUFFER_INFO in global var sinfo
//***************************************************************************
void console_init( const wchar_t* title )
{
  BOOL bSuccess;

  hStdOut = GetStdHandle( STD_OUTPUT_HANDLE );

  // Get screen information.
  // If this call fails, assume we are re-directing output.
  bSuccess = GetConsoleScreenBufferInfo( hStdOut, &sinfo );
  if (bSuccess == FALSE)
  {
    redirected = (GetFileType( hStdOut ) == FILE_TYPE_PIPE) ? 2 : 1;

    // Check if stderr is a console.
    hStdOut  = GetStdHandle( STD_ERROR_HANDLE );
    bSuccess = GetConsoleScreenBufferInfo( hStdOut, &sinfo );
    if (bSuccess == FALSE)
    {
      // Okay, it looks like we're not running in a console window.
      // Let's assume we're running in a UNIX-style terminal shell.
      // Distinguish between normal output and piped output by examining
      // the file index of stdout and stderr.
      BY_HANDLE_FILE_INFORMATION outfi, errfi;
      GetFileInformationByHandle( GetStdHandle( STD_OUTPUT_HANDLE ), &outfi );
      GetFileInformationByHandle( hStdOut, &errfi );
      if (outfi.nFileIndexLow  == errfi.nFileIndexLow &&
	  outfi.nFileIndexHigh == errfi.nFileIndexHigh)
      {
	redirected = 0;
      }
      hStdOut = 0;
      return;
    }
  }

  // Put up a meaningful console title.
  if (title)
    SetConsoleTitle( title );

  original_attribs = sinfo.wAttributes;
  GetConsoleMode( hStdOut, &original_mode );
  SetConsoleMode( hStdOut, (original_mode | ENABLE_PROCESSED_OUTPUT)
					  & ~ENABLE_WRAP_AT_EOL_OUTPUT );

  // Set up Ctrl-Break handler.
  SetConsoleCtrlHandler( (PHANDLER_ROUTINE)control_handler, TRUE );
}

//**********************************************************
void restore_console_attribs( void )
{
  if (hStdOut)
  {
    SetConsoleMode( hStdOut, original_mode );
    SetConsoleTextAttribute( hStdOut, original_attribs );
  }
}

//**********************************************************
WORD get_scode( void )
{
  WORD inchr;

  inchr = _getch();
  if (inchr == 0 || inchr == 0xE0)
    inchr |= _getch() << 8;

  return inchr;
}

//**********************************************************
int get_window_cols( void )
{
  const wchar_t* env;

  if (redirected)
    return reportcols;

  if (!hStdOut)
  {
    env = _wgetenv( L"COLUMNS" );
    return (env) ? wcstol( env, 0, 10 ) : reportcols;
  }

  // Use the window if the buffer is excessively long.
  return sinfo.dwSize.X <= MAX_PATH
	 ? sinfo.dwSize.X : sinfo.srWindow.Right - sinfo.srWindow.Left + 1;
}

//**********************************************************
int get_window_rows( void )
{
  const wchar_t* env;

  if (!hStdOut)
  {
    env = _wgetenv( L"LINES" );
    return (env) ? wcstol( env, 0, 10 ) : 25;
  }

  return sinfo.srWindow.Bottom - sinfo.srWindow.Top + 1;
}

//**********************************************************
void dgotoxy( int x, int y )
{
  curpos.X = x;
  curpos.Y = y;
  SetConsoleCursorPosition( hStdOut, curpos );
}

//**********************************************************
void set_text_attr( WORD tFGBG )
{
  SetConsoleTextAttribute( hStdOut, tFGBG );
}

//**********************************************************
void dreturn( void )
{
  DWORD wrlen;

  WriteConsole( hStdOut, L"\r", 1, &wrlen, 0 );
}

//**********************************************************
void dnewline( void )
{
  DWORD wrlen;

  // Avoid scrolling when line has already wrapped.
  if (!wrapped)
  {
    // Use the original color, so scrolling doesn't look weird.
    GetConsoleScreenBufferInfo( hStdOut, &sinfo );
    SetConsoleTextAttribute( hStdOut, original_attribs );
    WriteConsole( hStdOut, L"\n", 1, &wrlen, 0 );
    SetConsoleTextAttribute( hStdOut, sinfo.wAttributes );
  }
  wrapped = false;
}

//**********************************************************
void dclreol( void )
{
  int	slen;
  DWORD wrlen;

  GetConsoleScreenBufferInfo( hStdOut, &sinfo );
  slen = sinfo.dwSize.X - curpos.X;
  FillConsoleOutputCharacter( hStdOut, ' ',              slen, curpos, &wrlen );
  FillConsoleOutputAttribute( hStdOut, original_attribs, slen, curpos, &wrlen );
}

//**********************************************************
void dclrscr( void )
{
  COORD coord = { 0, 0 };
  int	slen;
  DWORD wrlen;

  slen = sinfo.dwSize.X * sinfo.dwSize.Y;
  FillConsoleOutputCharacter( hStdOut, ' ',              slen, coord, &wrlen );
  FillConsoleOutputAttribute( hStdOut, original_attribs, slen, coord, &wrlen );

  dgotoxy( 0, 0 ); // home the cursor
}

//**********************************************************
void dputc( wchar_t c )
{
  DWORD wrlen;

  if (c == '\n')
  {
    dnewline();
    return;
  }

  GetConsoleScreenBufferInfo( hStdOut, &sinfo );
  wrapped = false;
  WriteConsole( hStdOut, &c, 1, &wrlen, 0 );
  if (curpos.X == sinfo.dwSize.X - 1)
  {
    dnewline();
    wrapped = true;
  }
}

//**********************************************************
void dputnchar( wchar_t chr, unsigned char attr, int count )
{
  wchar_t buf[TEMPSTR_COUNT];

  set_text_attr( attr );
  if (count > 0)
  {
    wmemset( buf, chr, count );
    buf[count] = '\0';
    dputs( buf );
  }
}

//**********************************************************
void dputs( const wchar_t* outstr )
{
  const wchar_t* start;
  DWORD wrlen;
  int	x, max;

  GetConsoleScreenBufferInfo( hStdOut, &sinfo );
  max = sinfo.dwSize.X;
  x   = max - curpos.X;
  for (start = outstr; *outstr; ++outstr)
  {
    if (*outstr == '\n')
    {
      if (outstr > start)
      {
	wrapped = false;
	WriteConsole( hStdOut, start, outstr - start, &wrlen, 0 );
      }
      dnewline();
      x     = max;
      start = outstr + 1;
    }
    else if (--x == 0)
    {
      wrapped = false;
      WriteConsole( hStdOut, start, outstr + 1 - start, &wrlen, 0 );
      dnewline();
      wrapped = true;
      x       = max;
      start   = outstr + 1;
    }
  }
  if (outstr > start)
  {
    wrapped = false;
    WriteConsole( hStdOut, start, outstr - start, &wrlen, 0 );
  }
}
