//*********************************************************************
//
// DISKPARM.CPP: Display partition information about disk drive.
//
// Written by:   Daniel D. Miller
//		 Jason Hood
//
// Last update:  02/03/02 10:35
//		 28 Nov, 2012
//
//*********************************************************************
// Note on DISKPARM build:  The function WNetGetUniversalName()
// brings several problems with it.  First, it reportedly is not
// valid on Win95 drives.  Second, it requires MPR.LIB to be
// linked into the application (otherwise the compile fails).
//*********************************************************************

#include "ndata.hpp"
#include <winioctl.h>

wchar_t dpath[52*4] = L"d:\\";

static wchar_t vnbfr[MAXPATH];
static wchar_t fsnbfr[MAXPATH];

struct UNC
{
  wchar_t* uptr;
  wchar_t  ustr[MAXPATH];
};

//**********************************************************
void display_drive_summary( void )
{
  UINT	   dtype;
  wchar_t* drv;
  uchar    attr;

  // draw header
  nput_char( n.colornhead, ' ', (cols - 18) / 2 );
  nputs(     n.colornhead, L"Disk Drive Summary\n" );
  nput_line( n.colorframe, dline );
  nputs(     n.colornhead,
  L"   file sys         total space          free space  [Cluster Size] UNC path\n"
  L"   ========  ==================  ==================  =======================\n" );

  __int64 lfree  = 0;
  __int64 ltotal = 0;

  GetLogicalDriveStrings( sizeof(dpath)/sizeof(*dpath), dpath );
  for (drv = dpath; *drv; drv = wcschr( drv, '\0' ) + 1)
  {
    if (*drv == 'A' || *drv == 'B')     // skip the floppies (presumably)
      continue;

    if (!GetVolumeInformation(		// GetVolumeInformation(
	   drv, 			// LPCTSTR lpRootPathName,
	   vnbfr,			// LPTSTR  lpVolumeNameBuffer,
	   MAXPATH,			// DWORD    nVolumeNameSize,
	   NULL,			// LPDWORD lpVolumeSerialNumber,
	   NULL,			// LPDWORD lpMaximumComponentLength,
	   NULL,			// LPDWORD lpFileSystemFlags,
	   fsnbfr,			// LPTSTR  lpFileSystemNameBuffer,
	   MAXPATH			// DWORD    nFileSystemNameSize
	 ))				// );
      continue;

    dtype = GetDriveType( drv );
    attr  = (dtype == DRIVE_REMOTE) ? n.colordir :
	    (dtype == DRIVE_REMOVABLE || dtype == DRIVE_CDROM) ? n.colorsize :
	    n.colordefalt;

    nputf( attr, L"%c: %-8s  ", drv[0], fsnbfr );
    if (wcslen( fsnbfr ) > 8)
    {
      ncrlf();
      nput_char( attr, ' ', 2 + 1 + 8 + 2 );
    }

    unsigned __int64 freec1, frees1, totals1;
    if (!GetDiskFreeSpaceEx( drv,
			     (ULARGE_INTEGER*)&freec1,
			     (ULARGE_INTEGER*)&totals1,
			     (ULARGE_INTEGER*)&frees1 ))
    {
      nput_char( attr, ' ', 18 + 2 + 18 + 2 );
      totals1 = frees1 = 0;
    }
    else
    {
      nputf( attr, L"%18s  %18s  ", i64toss( totals1 ), i64toss( frees1 ) );
    }

    // if network, pull the UNC name, otherwise show drive type
    if (dtype == DRIVE_REMOTE)
    {
      UNC UNCpaths;
      UNCpaths.uptr = UNCpaths.ustr;
      DWORD bufsize = MAXPATH;

      nputs( attr, (WNetGetUniversalName( drv, UNIVERSAL_NAME_INFO_LEVEL,
					  &UNCpaths, &bufsize ) == NO_ERROR)
		   ? UNCpaths.uptr : L"Network Drive" );
    }
    else
    {
      get_disk_info( drv );
      nputf( attr, (n.hex) ? L"[%5X] %s" : L"[%5u] %s",
		   (unsigned)clbytes, vnbfr );

      //if (dtype == DRIVE_FIXED)
      {
	lfree  += frees1;
	ltotal += totals1;
      }
    }
    ncrlf();
  }

  // display drive summary
  nput_line( n.colorframe, dline );
  nputf( n.colorxhead, L"             %18s  %18s",
		       i64toss( ltotal ), i64toss( lfree ) );
  nputs( n.colornhead, L"  Total Physical space\n" );
}

//*****************************************************************
// this function determines the following information:
// - What file system is in use??
// - Total partition size in bytes
// - Free partition space in bytes
// - Cluster size for active partition
// - Volume label for requested partition
//*****************************************************************
void get_disk_info( const wchar_t* dstr )
{
  if (!GetVolumePathName( dstr, dpath, sizeof(dpath) ))
    *dpath = *dstr;

  if (!GetVolumeInformation( dpath,	  // LPCTSTR lpRootPathName,
			     volume_name, // LPTSTR  lpVolumeNameBuffer,
			     MAXPATH,	  // DWORD    nVolumeNameSize,
			     NULL,	  // LPDWORD lpVolumeSerialNumber,
			     NULL,	  // LPDWORD lpMaximumComponentLength,
			     NULL,	  // LPDWORD lpFileSystemFlags,
			     fsnbfr,	  // LPTSTR  lpFileSystemNameBuffer,
			     MAXPATH	  // DWORD    nFileSystemNameSize
			   ))
  {
    *volume_name = '\0';	// try to keep going...
    is_ntfs = false;
  }
  else
    is_ntfs = (wcsncmp( fsnbfr, L"NTFS", 4 ) == 0);

  if (*volume_name == '\0')
    wcscpy( volume_name, L"undefined" );

  unsigned __int64 freec1, frees1, totals1;
  if (!GetDiskFreeSpaceEx( dstr,
			   (ULARGE_INTEGER*)&freec1,
			   (ULARGE_INTEGER*)&totals1,
			   (ULARGE_INTEGER*)&frees1 ))
  {
    // emergency handling when this function fails
    diskbytes = 0;
    diskfree = 0;
    clbytes = 4096;
  }
  // if GetDiskFreeSpaceEx succeeds, proceed normally
  else
  {
    DWORD spc, bps;
    diskbytes = totals1;
    diskfree = frees1;
    clbytes = (GetDiskFreeSpace( dstr, &spc, &bps, NULL, NULL ))
	      ? spc * bps : 4096;	// might be wrong in 9X
  }
}


//*****************************************************************
__int64 get_disk_size( const wchar_t* name, __int64 fsize, BOOL compressed )
{
  // a small file on NTFS may fit within the MFT
  if (is_ntfs && fsize < 768)
  {
    HANDLE hFile;
    STARTING_VCN_INPUT_BUFFER vcn;
    RETRIEVAL_POINTERS_BUFFER lcn;
    DWORD ret;
    BOOL  ok;

    hFile = CreateFile( name, 0, FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, 0, 0 );
    if (hFile != INVALID_HANDLE_VALUE)
    {
      vcn.StartingVcn.QuadPart = 0;
      ok = DeviceIoControl( hFile, FSCTL_GET_RETRIEVAL_POINTERS,
			    &vcn, sizeof(vcn), &lcn, sizeof(lcn), &ret, NULL );
      CloseHandle(hFile);
      if (!ok && GetLastError() == ERROR_HANDLE_EOF)
	return 0;
    }
  }
  if (compressed)
  {
    LARGE_INTEGER uli;
    uli.LowPart = GetCompressedFileSize( name, (PDWORD)&uli.HighPart );
    if (uli.LowPart != INVALID_FILE_SIZE || GetLastError() == NO_ERROR)
      return uli.QuadPart;
  }
  return ((fsize - 1) | (clbytes - 1)) + 1;
}
