//*****************************************************************
//
// FILEREAD.CPP: NDIR file-reading routines
//
// Written by:	 Daniel D. Miller  (the Derelict)
//		 Jason Hood
//
// Last update:  01/15/01 21:20PM
//		 19 Aug, 2013
//
//*****************************************************************

#include "ndata.hpp"

// from nsort.cpp
extern int sort_name( const ffdata* a, const ffdata* b );
extern void sort_filelist( void );

// from filelist.cpp
extern void filesummary( void );
extern int total_filecount;

//*****************  function prototypes  *****************
static void read_files( void );

static uchar fclr;
static DWORD fid;
static int   dir_tally; // keep track of subdirectories for recursion

//***************************************************************
static void getcolor( const wchar_t* filename, const wchar_t* ext )
{
  attrib_list* aptr;

  if (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
  {
    fclr = n.colordir;
    fid  = ('d' << 16) | ('i' << 8) | 'r';
    return;
  }

  // search the filename list
  aptr = attr_table[which_attrib_table( filename )];
  for (; aptr; aptr = aptr->next)
    if (_wcsicmp( aptr->ext, filename ) == 0)
      goto found;

  // search the extension list
  aptr = attr_table[which_attrib_table( ext )];
  for (; aptr; aptr = aptr->next)
    if (_wcsicmp( aptr->ext, ext ) == 0)
      goto found;

  // search the filename wildcard list
  for (aptr = attr_table[27]; aptr; aptr = aptr->next)
    if (wildcard( aptr->ext, filename ))
      goto found;

  // search the extension wildcard list
  for (aptr = attr_table[55]; aptr; aptr = aptr->next)
    if (wildcard( aptr->ext, ext ))
      goto found;

  // if not found, assign default color
  fclr = n.colordefalt;
  fid  = ('d' << 24) | ('e' << 16) | ('f' << 8) | 'c';
  return;

found:
  fclr = aptr->attr;
  fid  = aptr->id;
}

//*********************************************************
static int dirs_cmp( const void* a, const void* b )
{
  ffdata af, bf;
  af.filename = *(wchar_t**)a;
  bf.filename = *(wchar_t**)b;
  return sort_name( &af, &bf );
}

//*********************************************************
// File-listing routine
//*********************************************************
void file_listing( void )
{
  //**************************************************
  // initialize file pointer and filecount,
  // in case of multiple filespecs.
  //**************************************************
  while (ftop)
  {
    ffdata* fkill = ftop;
    ftop = ftop->next;
    delete fkill->filename;
    delete fkill;
  }
  filecount = dirscount = filesdisp = 0;

  read_files();

  //***********************************************
  // Sort directory data
  //***********************************************
  if (filecount > 1)
    sort_filelist();

  // restrict the count to that requested
  if (filesmax && filecount > abs(filesmax))
  {
    int files = filesmax;
    if (files < 0)
    {
      // delete files from the front so as to leave filesmax at the end
      files += filecount;
      filecount -= files;
      do
      {
	ffdata* fkill = ftop;
	ftop = ftop->next;
	delete fkill->filename;
	delete fkill;
      } while (--files);
    }
    else
    {
      // skip the first filesmax files, deleting the rest
      filecount = files;
      ffdata* fkill = ftop;
      ffdata* fprev;
      do
      {
	fprev = fkill;
	fkill = fkill->next;
      } while (--files);
      fprev->next = NULL;
      while (fkill)
      {
	ffdata* ftemp = fkill;
	fkill = fkill->next;
	delete ftemp->filename;
	delete ftemp;
      }
    }
  }

  // now do the file-listing...
  if (filecount > 0 || !n.recurse)
  {
    if (total_filecount != 0 && !n.batch)
      ncrlf();
    display_files();
  }

  //*********************************************************
  if (n.recurse)
  {
    static int recurse;
    HANDLE handle;
    wchar_t** dir;
    int dirs;
    base_path[base_len] = '*';
    base_path[base_len+1] = '\0';
    handle = FindFirstFile( base_path, &fdata );
    if (handle == INVALID_HANDLE_VALUE)
    {
      base_path[base_len] = '\0';
      return;
    }
    ++recurse;
    dir = new wchar_t*[dir_tally];
    if (!dir)
      error_exit( OUT_OF_MEMORY, NULL );
    dirs = 0;
    for (;;)
    {
      if ((fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) &&
	  (n.show_all || !(fdata.dwFileAttributes & (FILE_ATTRIBUTE_HIDDEN|
						     FILE_ATTRIBUTE_SYSTEM)))
	  && ! ((fdata.cFileName[0] == '.' && (fdata.cFileName[1] == '\0' ||
		(fdata.cFileName[1] == '.' && fdata.cFileName[2] == '\0')))))
      {
	dir[dirs] = _wcsdup( fdata.cFileName );
	if (!dir[dirs++])
	  error_exit( OUT_OF_MEMORY, NULL );
      }
      while (!FindNextFile( handle, &fdata ))
	if (GetLastError() != ERROR_ACCESS_DENIED)
	  goto done;
    } done:
    FindClose( handle );
    if (dirs)
    {
      qsort( dir, dirs, sizeof(wchar_t*), dirs_cmp );
      for (int i = 0; i < dirs; ++i)
      {
	wcscpy( base_path + base_len, dir[i] );
	base_len += wcslen( dir[i] );
	base_path[base_len++] = '\\';
	file_listing();
	base_len -= 1 + wcslen( dir[i] );
	free( dir[i] );
      }
    }
    delete[] dir;
    base_path[base_len] = '\0';
    if (--recurse == 0 && !n.batch)
      filesummary();
  }
}

//*********************************************************
// This loops thru all files in one subdirectory,
// calling update_filelist() to add files to the
// global linked list.
//*********************************************************
static void read_files( void )
{
  int	  len, nlen;
  wchar_t *strptr, *extptr;
  ffdata* ftemp;

  base_path[base_len]	= '*';
  base_path[base_len+1] = '\0';

  if (!find_first( base_path ))
  {
    base_path[base_len] = '\0';
    return;
  }

  // loop on find_next
  do
  {
    strptr = (n.lfn_off && *fdata.cAlternateFileName)
	     ? fdata.cAlternateFileName : fdata.cFileName;
    extptr = wcsrchr( strptr, '.' );
    len    = wcslen( strptr );
    nlen   = (extptr) ? (int)(extptr - strptr) : len;

    // Verify we have an actual short name, which GetShortPathName does not
    // actually do.  This really only needs to be done on NTFS drives.
    if (n.lfn_off)
    {
      if (!*fdata.cAlternateFileName)
      {
	if (nlen > 8 || len - nlen > 4 ||
	    extptr == strptr || (extptr && wcschr( strptr, '.' ) != extptr))
	{
	  continue;
	}
      }
      // convert filename to lower case if appropriate
      if (!n.ucase)
	_wcslwr( strptr );
    }

    //****************************************************
    // allocate and initialize the structure
    //****************************************************
    ftemp = new ffdata;
    if (!ftemp)
      error_exit( OUT_OF_MEMORY, NULL );

    ftemp->filename = new wchar_t[len + 1 + nlen + 1];
    if (!ftemp->filename)
      error_exit( OUT_OF_MEMORY, NULL );

    wmemcpy( ftemp->filename, strptr, len + 1 );
    ftemp->name = ftemp->filename + len + 1;
    wmemcpy( ftemp->name, strptr, nlen );
    ftemp->name[nlen] = '\0';
    ftemp->ext = ftemp->filename + nlen;

    ftemp->attrib  = (uchar)fdata.dwFileAttributes;
    ftemp->dirflag = ftemp->attrib & FILE_ATTRIBUTE_DIRECTORY;
    ftemp->compflag = (fdata.dwFileAttributes & (FILE_ATTRIBUTE_COMPRESSED|
						 FILE_ATTRIBUTE_SPARSE_FILE))
		      ? true : false;

    ftemp->ft = (n.fdate_option == FDATE_LAST_ACCESS) ? fdata.ftLastAccessTime :
		(n.fdate_option == FDATE_CREATE_TIME) ? fdata.ftCreationTime :
		fdata.ftLastWriteTime;

    // convert file size
    i64tol iconv;
    iconv.u[0]	 = fdata.nFileSizeLow;
    iconv.u[1]	 = fdata.nFileSizeHigh;
    ftemp->fsize = iconv.i;

    // look up color in table
    if (n.color)
    {
      if (!ccount)
	getcolor( ftemp->filename, ftemp->ext );
      ftemp->color = fclr;
    }

    //****************************************************
    // add the structure to the file list
    //****************************************************
    if (!ftop)
      ftop = ftemp;
    else
      ftail->next = ftemp;
    ftail = ftemp;
    ftemp->next = 0;

    filecount++;
    if (ftemp->dirflag)
      dirscount++;

    // search for another file
  } while (find_next());

  base_path[base_len] = '\0';
}

//*********************************************************
static bool find_match( void )
{
  // sloppy programming in wildcard() prevents using this directly
  static wchar_t exes[] = L"*.exe;*.com;*.bat;*.cmd;*.btm";

  // filter out directories if not requested
  if (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
  {
    // filter out "." and ".."
    if (fdata.cFileName[0] == '.' && (fdata.cFileName[1] == '\0' ||
	(fdata.cFileName[1] == '.' && fdata.cFileName[2] == '\0')))
      return false;
    ++dir_tally;
    if (n.tree)
      return true;
    if (n.files_only)
      return false;
  }

  // match the appropriate attributes
  if (n.show_all & 2)
  {
    if ((fdata.dwFileAttributes & attrib_excl) ||
	(fdata.dwFileAttributes & attrib_must) != attrib_must ||
	(!(fdata.dwFileAttributes & attrib_have)))
      return false;
  }
  else if (!n.show_all)
  {
    if (fdata.dwFileAttributes & (FILE_ATTRIBUTE_HIDDEN|FILE_ATTRIBUTE_SYSTEM))
      return false;
  }

  if (fdata.dwFileAttributes == FILE_ATTRIBUTE_NORMAL)
    fdata.dwFileAttributes = 0;

  bool fn_okay = false;
  for (int i = start; i < finish; i++)
  {
    if (wildcard( target[i] + wild_pos, fdata.cFileName ))
    {
      fn_okay = true;
      break;
    }
  }
  if (!fn_okay)
    return false;

  if (n.exec_only)
    if (!wildcard( exes, fdata.cFileName ))
      return false;

  if (ccount)
  {
    const wchar_t* ext = wcsrchr( fdata.cFileName, '.' );
    if (!ext)
      ext = L"";
    getcolor( fdata.cFileName, ext );
    for (int i = 0; i < ccount; ++i)
      if (fclr == targetclr[i] || fid == targetclr[i])
	return true;

    return false;
  }

  return true;
}

//*********************************************************
static HANDLE handle;

bool find_first( const wchar_t* path )
{
  dir_tally = 0;
  handle = FindFirstFile( path, &fdata );
  if (handle == INVALID_HANDLE_VALUE)
    return false;

  return (find_match()) ? true : find_next();
}

bool find_next( void )
{
  do
  {
    while (FindNextFile( handle, &fdata ))
      if (find_match())
	return true;
  } while (GetLastError() == ERROR_ACCESS_DENIED);

  FindClose( handle );
  return false;
}
