/*
 EPSHeader

   File: filmatch.c
   Author: J. Kercheval
   Created: Thu, 03/14/1991  22:22:01
*/

/*
 EPSRevision History

   J. Kercheval  Wed, 02/20/1991  22:29:01  Released to Public Domain
   J. Kercheval  Fri, 02/22/1991  15:29:01  fix '\' bugs (two :( of them)
   J. Kercheval  Sun, 03/10/1991  19:31:29  add error return to matche()
   J. Kercheval  Sun, 03/10/1991  20:11:11  add is_valid_pattern code
   J. Kercheval  Sun, 03/10/1991  20:37:11  beef up main()
   J. Kercheval  Tue, 03/12/1991  22:25:10  Released as V1.1 to Public Domain
   J. Kercheval  Thu, 03/14/1991  22:22:25  remove '\' for DOS file parsing
   J. Kercheval  Thu, 03/28/1991  20:58:27  include filmatch.h

   Jason Hood, 1 August, 1997 - removed error testing in matche()
	       4 August, 1997 - case testing
				renamed match() to wildcard()
	      10 August, 1997 - case testing in [] (oops)
	      31 July, 2005   - match multiple patterns and exclusions
	      9 December, 2005 - remove backslash escaping in range
				 allow ']' as end of range
				 just use '!' as not in set.
*/

/*
   Wildcard Pattern Matching
*/


#include "filmatch.h"
#include <ctype.h>
#include <string.h>


int matche_after_star( const wchar_t*, const wchar_t* );


/*----------------------------------------------------------------------------
*
* Return TRUE if PATTERN has any special wildcard characters
*
----------------------------------------------------------------------------*/

BOOLEAN is_pattern( const wchar_t* p )
{
  const wchar_t* s;

  for (s = p; *s; ++s)
  {
    if (*s == '/' || *s == '\\')
      p = s + 1;
  }

  while (*p)
  {
    switch (*p++)
    {
      case '!':
	if (*p == '\0')
	  break;
      case '*':
      case '?':
      case '[':
      case ';':
	return TRUE;
    }
  }
  return FALSE;
}


/*----------------------------------------------------------------------------
*
* Return TRUE if PATTERN is a well formed regular expression according
* to the above syntax
*
*   TRUE  - pattern is well formed
*   FALSE - [..] construct has no end bracket (ie "[abc-g")
*
----------------------------------------------------------------------------*/

BOOLEAN is_valid_pattern( const wchar_t* p )
{
  /* loop through pattern to EOS */
  for (; *p; ++p)
  {
    if (*p == '[')
    {
      /* the [..] construct must be well formed */
      if (*++p == '!' && p[1] != ']')
	++p;

      /* loop to end of [..] construct */
      do
      {
	if (!*p)
	  return FALSE;

	/* if this a range */
	if (*++p == '-' && *++p)
	  ++p;
      } while (*p != ']');
    }
  }

  return TRUE;
}


/*----------------------------------------------------------------------------
*
*  Match the pattern PATTERN against the string TEXT;
*
*  returns MATCH_VALID if pattern matches, or an errorcode as follows
*  otherwise:
*
*            MATCH_PATTERN  - bad pattern
*            MATCH_RANGE    - match failure on [..] construct
*            MATCH_ABORT    - premature end of text string
*            MATCH_END      - premature end of pattern string
*            MATCH_VALID    - valid match
*
*
*  A match means the entire string TEXT is used up in matching.
*
*  In the pattern string:
*       `*' matches any sequence of characters (zero or more)
*       `?' matches any character
*       [SET] matches any character in the specified set,
*	[!SET] matches any character not in the specified set.
*
*  A set is composed of characters or ranges; a range looks like
*  character hyphen character (as in 0-9 or A-Z).  [0-9a-zA-Z_] is the
*  minimal set of characters allowed in the [..] pattern construct.
*  Other characters are allowed (ie. 8 bit characters) if your system
*  will support them.
*
*  To match '-' in a set, place it first or after a range; to match ']'
*  place it first; '!' will match literally anywhere but first, or if
*  used alone.
*
----------------------------------------------------------------------------*/

int matche( const wchar_t* p, const wchar_t* t )
{
  wchar_t range_start, range_end;	/* start and end of range */

  BOOLEAN invert;		/* is this [..] or [!..] */
  BOOLEAN member_match; 	/* have I matched the [..] construct? */

  for (; *p; p++, t++)
  {
    /* if this is the end of the text then this is the end of the match */
    if (!*t)
      return (*p == '*' && *++p == '\0') ? MATCH_VALID : MATCH_ABORT;

    /* determine and react to pattern type */
    switch (*p)
    {
      /* single any character match */
      case '?':
      break;

      /* multiple any character match */
      case '*':
	return matche_after_star( p, t );

      /* [..] construct, single member/exclusion character match */
      case '[':
	/* move to beginning of range */
	p++;

	/* check if this is a member match or exclusion match */
	invert = FALSE;
	if (*p == '!' && p[1] != ']')
	{
	  invert = TRUE;
	  p++;
	}

	member_match = FALSE;

	do
	{
	  range_start = range_end = *p++;

	  /* check for range bar */
	  if (*p == '-')
	  {
	    /* get the range end */
	    range_end = *++p;
	    ++p;
	  }

	  range_start = towlower( range_start );
	  range_end   = towlower( range_end );

	  /* make sure the range letters have the proper
	     relationship to one another before comparison */
	  if (range_start > range_end)
	  {
	    wchar_t temp = range_start;
	    range_start  = range_end;
	    range_end	 = temp;
	  }

	  /* if the text character is in range then match found. */
	  if (towlower( *t ) >= range_start &&
	      towlower( *t ) <= range_end)
	  {
	    member_match = TRUE;
	    break;
	  }

	  /* if end of construct then loop is done */
	} while (*p != ']');

	/* if there was a match in an exclusion set then no match */
	/* if there was no match in a member set then no match */
	if ((invert && member_match) ||
	    !(invert || member_match))
	  return MATCH_RANGE;

	/* if this is not an exclusion then skip the rest of the [...]
	    construct that already matched. */
	while (*p != ']')
	{
	  if (*++p == '-')
	    p += 2;
	}
      break;

       /* must match this character exactly */
      default:
	if (towlower( *p ) != towlower( *t ))
	  return MATCH_LITERAL;
    }
  }

  /* if end of text not reached then the pattern fails */
  return (*t) ? MATCH_END : MATCH_VALID;
}


/*----------------------------------------------------------------------------
*
* recursively call matche() with final segment of PATTERN and of TEXT.
*
----------------------------------------------------------------------------*/

int matche_after_star( const wchar_t* p, const wchar_t* t )
{
  int	  match = 0;
  wchar_t nextp;

  /* pass over existing ? and * in pattern */
  while (*p == '?' || *p == '*')
  {
    /* take one char for each ? */
    if (*p == '?')
    {
      /* if end of text then no match */
      if (!*t++)
	return MATCH_ABORT;
    }

    /* move to next char in pattern */
    p++;
  }

  /* if end of pattern we have matched regardless of text left */
  if (!*p)
    return MATCH_VALID;

  /* get the next character to match which must be a literal or '[' */
  nextp = towlower( *p );

  /* Continue until we run out of text or definite result seen */
  do
  {
    /* a precondition for matching is that the next character
       in the pattern match the next character in the text or that
       the next pattern char is the beginning of a range.  Increment
       text pointer as we go here */
    if (nextp == towlower( *t ) || nextp == '[')
      match = matche( p, t );

    /* if the end of text is reached then no match */
    if (!*t++) match = MATCH_ABORT;

  } while (match != MATCH_VALID &&
	   match != MATCH_ABORT);

  /* return result */
  return match;
}


/*----------------------------------------------------------------------------
*
* wildcard() is a shell to matche() to return only BOOLEAN values.
*
* jmh 050731: allow it to match multiple patterns, separated by ';', and
*              exclusions, separated by '!'.  Use the set notation to treat
*              the character literally.
*             eg: "!*.exe;*.com" will match all files except .EXE & .COM.
*                 "!*.*" will match files with no extension.
*                 "*.txt;*.doc!read*" will match all .TXT & .DOC files,
*                 except those starting with READ.
*		  "[!]*" will match all files starting with '!'.
* jmh 060401: ignore the path, only testing the name and treat trailing '!'
*	       as a literal.
*
----------------------------------------------------------------------------*/

BOOLEAN wildcard( wchar_t* p, const wchar_t* t )
{
  int	   error_type;
  wchar_t* alt; 		// alternative
  wchar_t* exc; 		// exclusion
  wchar_t  falt, fexc;		// found ...
  BOOLEAN  inverted = FALSE;	// matching exclusion rather than pattern
  int	   rc;

  for (alt = p; *alt; ++alt)
  {
    if (*alt == '/' || *alt == '\\')
      p = alt + 1;
  }

  if (!is_valid_pattern( p ))
    return FALSE;

  for (exc = p; *exc && *exc != '!'; ++exc)
  {
    if (*exc == '[')
    {
      do
      {
	if (*++exc == '-')
	  exc += 2;
      } while (*exc != ']');
    }
  }
  fexc = *exc;
  if (*exc == '!' && exc[1])
    *exc = '\0';
  if (exc == p && fexc)
    p = L"*";
  rc = FALSE;
  do
  {
    for (alt = p; *alt && *alt != ';'; ++alt)
    {
      if (*alt == '[')
      {
	do
	{
	  if (*++alt == '-')
	    alt += 2;
	} while (*alt != ']');
      }
    }
    falt = *alt;
    if (falt)			// prevent modifying the static string
      *alt = '\0';
    error_type = matche( p, t );
    if (falt)
      *alt = falt;
    if (error_type == MATCH_VALID)
    {
      if (inverted)
      {
	rc = FALSE;
	break;
      }
      rc = TRUE;
      if (!fexc)
	break;
      inverted = TRUE;
      p = exc + 1;
    }
    else
    {
      if (!falt)
	break;
      p = alt + 1;
    }
  } while (*p);

  *exc = fexc;

  return rc;
}
