//*****************************************************************
//
// jNDIR.CPP:	The Ultimate directory program (32-bit).
//
// Written by:	Daniel D. Miller  (the derelict)
//		Jason Hood
//
// Version 1.00, 28 Dec, 2005
// - clean slate
//
// 1 Apr, 2006
// - fixed /kID
// - ignore the path when testing for a pattern
// - treat trailing '!' as literal
//
// 25 Apr, 2006
// - tweaked numeric sorting and added option/config to toggle it
//
// 8 Jan, 2007
// - fixed infinite loop caused by corrupted directories
//
// 10 Sep, 2011
// - fixed excluded attribute matching when no attributes present
// - added /v option to limit the number of files displayed
//
// Version 1.01, 11 Nov, 2011
// - minor changes to prevent warnings in gcc 4.6.1 (but still
//   released using gcc 3.4.5, being 32K smaller)
// - comment out cols and rows when writing the default ndir.ini
//
// Version 1.02, 6 February, 2012
// - tree: adjust size so as to use the whole width
// - tree: adjust size to preserve alignment for really deep paths
// - tree: fix display error for really deep paths
// - tree: show "Error!" if a directory is inaccessible
//
// Version 1.10, 28 to 30 November, 2012
// - convert command line file specs to OEM
// - allow more than 20 file specs
// - allow formats to be defined in ndir.ini
// - new ini option one_level to set /, option
// - use the path when getting disk info, to support NTFS mount points
// - recognise NTFS compressed/sparse/resident files for disk space calc
// - sorting by name uses a word/locale sort
// - ignore hexadecimal numbers when sorting by number
// - tree: always find directories, even with non-directory options
// - tree: do not include an initial cluster for NTFS directories
// - added /. option to recurse into subdirectories
// - added /$ option to use hexadecimal for sizes
// - added /mN option to force the maximum number of columns
// - added /5 option to list file & age
// - added #l/#L formats for age
// - display separate file and directory counts
// - tweaked output to allow terabyte sizes
// - improved the ANSI output
// - fix default length for size format in batch mode
// - batch mode #f/#q will include the original path
// - use the default color to blank out columns, not the frame color
// - remove the trailing column separator on minimized single-row listings
// - add the trailing column separators for horizontal multi-row listing
//
// Version 1.11, 15 August, 2013
// - /m may use megabyte disk sizes to avoid writing an extra line
// - disable Wow64 file system redirection
// - restore the console title (redirection didn't do it automatically)
// - fix /. in root directory (don't skip first directory)
// - fix /. pause test
// - fix /d, total pause count
// - fix /b original path
// - minor changes to compile with VC
//
// Version 1.12, 19 August, 2013
// - apply /vN to each directory
// - -#N uses equivalent -N format (if N not explicitly defined)
// - fix age in batch format
// - remove blank line between directories in batch format
// - recognise --help & --version
//
// Version 1.13, 1 September, 2013
// - don't change console title if redirected
//
// Version 1.14, 21 September, 2013
// - fix batch mode paths (didn't sort target_path with target)
//
// Version 1.20, 25 & 27 August, 2014
// - stricter requirements for numeric sorting
// - sorting by name will first sort up to the first dot
// - switch to Unicode (configuration and HTML are UTF-8; ID remains ASCII)
// - allow /j for graphic characters when redirecting
//
// Version 1.21, 16 September, 2014
// - if numbers sort the same, test digits; if still the same, look for more
//
// Version 1.22, 18 & 19 September, 2014:
// - improvements to the tree
// - double up the file name buffer, to allow for longer-than-usual names
//
// Version 1.23, 26 September, 2014:
// - restore console wrapping for /b and help
//
// Version 1.24, 12 January, 2015:
// - use window width if buffer width is larger than MAX_PATH
// - use TEMPSTR_COUNT, not 500
// - change /cN to 10 to TEMPSTR_COUNT-2, reducing larger values
//
//*****************************************************************

#include "ndata.hpp"
#include "conio32.hpp"
#include <io.h>
#include <locale.h>

#ifdef __MINGW32__
int _CRT_glob = 0;
#endif

const wchar_t Version[] = L"jNDIR.EXE, Version " PVER;

// from QUALIFY.CPP
extern unsigned qualify( wchar_t* argptr );

/***********************  qualify() constants  ************************/
#define QUAL_WILDCARDS	0x01
#define QUAL_NO_PATH	0x02
#define QUAL_IS_FILE	0x04
#define QUAL_INV_DRIVE	0x80

// from CMD_LINE.CPP
extern void parse_command_args( int start, int argc, wchar_t** argv );

extern void output_html_header( const wchar_t* title );
extern void output_html_footer( void );

extern const wchar_t* fprint[6][2];

//***************  function prototypes  ***************
static void process_filespecs( void );
static void sort_target_paths( void );
static void read_config_file( void );
static void parse_color_entry( wchar_t* iniptr );

//***********************************************************
void insert_target_filespec( const wchar_t* fstr )
{
  wchar_t buf[MAXPATH];

  wcscpy( buf, fstr );
  if (qualify( buf ) == QUAL_INV_DRIVE)
    error_exit( INV_DRIVE, buf );

  if (tcount == tcapacity)
  {
    wchar_t** t = (wchar_t**)realloc( target, (tcapacity + 20) * sizeof(wchar_t*) );
    if (!t)
      error_exit( OUT_OF_MEMORY, NULL );
    target = t;
    t = (wchar_t**)realloc( target_path, (tcapacity + 20) * sizeof(wchar_t*) );
    if (!t)
      error_exit( OUT_OF_MEMORY, NULL );
    target_path = t;
    tcapacity += 20;
  }
  target[tcount] = _wcsdup( buf );
  if (!target[tcount])
    error_exit( OUT_OF_MEMORY, NULL );

  // Make a copy of the original path, for batch mode.
  int len = wcslen( buf );
  bool dir = (len >= 2 && buf[len-1] == '*' && buf[len-2] == '\\');
  wcscpy( buf, fstr );
  len = wcslen( buf );
  if (dir && len >= 2 && (buf[len-1] != '*' ||
      (buf[len-2] != '\\' && buf[len-2] != '/')))
  {
    // Just a path was given, so explicitly add the backslash.
    if (buf[len-1] != '/' && buf[len-1] != '\\' && buf[len-1] != ':')
    {
      buf[len++] = '\\';
      buf[len] = '\0';
    }
  }
  else
  {
    while (--len >= 0 && buf[len] != '/' && buf[len] != '\\')
      ;
    if (len < 0 && buf[1] == ':')
      len = 1;
    buf[len+1] = '\0';
  }
  target_path[tcount] = _wcsdup( buf );
  if (!target_path[tcount])
    error_exit( OUT_OF_MEMORY, NULL );

  tcount++;
}

//*****************************************************************
static const wchar_t local_ini_name[] = L"ndir.ini";
static wchar_t ininame[MAXPATH];

#if defined(__MINGW32__) && !defined(__MINGW64_VERSION_MAJOR)
// Thanks to Coder for Life.
// http://www.coderforlife.com/projects/utilities/
int wmain( int, wchar_t** );
extern "C" void __wgetmainargs( int*, wchar_t***, wchar_t***, int, int* );
int main()
{
  wchar_t **argv, **envp;
  int argc, si = 0;
  __wgetmainargs( &argc, &argv, &envp, 0, &si );
  return wmain( argc, argv );
}
#endif

int wmain( int argc, wchar_t** argv )
{
  // Set the locale code page so wide-string conversions work as expected.
  wchar_t cp[16];
  _snwprintf( cp, 16, L".%u", GetConsoleOutputCP() );
  _wsetlocale( LC_CTYPE, cp );

  console_init( _isatty( 1 ) ? Version : NULL );

  typedef BOOL (WINAPI *TFsRedir)( PVOID* );
  TFsRedir DisableWow64FsRedirection =
		  (TFsRedir)GetProcAddress( GetModuleHandle( L"kernel32.dll" ),
					    "Wow64DisableWow64FsRedirection" );
  if (DisableWow64FsRedirection)
  {
    PVOID val;
    DisableWow64FsRedirection( &val );
  }

  //***********************************************************
  // Check for NDIR environment variable
  //***********************************************************

  int startIdx = 1;

  *ininame = '\0';      // on the very odd chance GetModuleFileName fails
  GetModuleFileName( NULL, ininame, MAXPATH );
  wchar_t* strptr = wcsrchr( ininame, '\\' );
  if (!strptr)
    strptr = ininame;
  else
    ++strptr;
  wcscpy( strptr, local_ini_name );

  wchar_t* options = _wgetenv( L"NDIR" );
  if (options)
  {
    argv[0]  = options;
    startIdx = 0;
  }

  //***********************************************************
  // first read default settings
  //***********************************************************
  read_config_file();

  //***********************************************************
  // override defaults with command line and environment vars
  //***********************************************************
  parse_command_args( startIdx, argc, argv );
  verify_flags();

  //***********************************************************
  // Execute the requested command
  //***********************************************************
  output_html_header( Version );
  display_logo();

  if (n.help)
  {
    info();
  }
  else if (n.drive_summary)
  {
    display_drive_summary();
  }
  else
  {
    // If no filespec was given, use current path
    if (tcount == 0)
      insert_target_filespec( L"." );

    sort_target_paths();
    process_filespecs();
  }

  output_html_footer();

  error_exit( DATA_OKAY, NULL );
  return 0;
}

/**********************************************************************/
/**                     File listing routines                        **/
/**********************************************************************/
static void process_filespecs( void )
{
  wchar_t* strptr;

  /********************************************************************/
  /***********************  loop on filespecs  ************************/
  /********************************************************************/

  // check for validity of long_filename functions
  lfn_supported = 1 - n.lfn_off;

  start = finish = 0;
  while (1)
  {
    // Extract base path from first filespec,
    // and strip off filename
    wcscpy( base_path, target[start] );
    strptr  = wcsrchr( base_path, '\\' ) + 1;
    *strptr = '\0';     // skip past backslash and strip off filename

    base_len = wild_pos = strptr - base_path;

    //**************************************************
    get_disk_info( base_path );

    // seek out all other filespecs with same path
    while (++finish < tcount)
    {
      // strip filename from next argument
      strptr = wcsrchr( target[finish], '\\' ) + 1;

      // now see if they are equal
      if (strptr - target[finish] != base_len ||
	  _wcsnicmp( base_path, target[finish], base_len ) != 0)
	break;
    }

    if (n.lfn_off)
    {
      base_len = GetShortPathName( base_path, base_path, MAXPATH );
      if (base_len == 0)
	base_len = wild_pos;
    }

    if (n.tree)
      tree_listing();
    else
      file_listing();

    start = finish;
    if (start == tcount)
      break;
    if (!n.batch)
      ncrlf();
  } // while not done
}

//**************************************************
// Sort filespecs alphabetically,
// to group those in one directory.
//**************************************************
static void sort_target_paths( void )
{
  wchar_t* strptr;
  int	   i, j, k;
  int	   rc;

  for (j = 0; j < tcount - 1; j++)
  {
    for (i = j + 1; i < tcount; i++)
    {
      rc = _wcsicmp( target[j], target[i] );
      if (rc > 0)
      {
	strptr	  = target[i];
	target[i] = target[j];
	target[j] = strptr;
	strptr	  = target_path[i];
	target_path[i] = target_path[j];
	target_path[j] = strptr;
      }
      else if (rc == 0)
      {
	// Delete duplicated filespecs.
	free( target[i] );
	free( target_path[i] );
	for (k = i + 1; k < tcount; k++)
	{
	  target[k-1] = target[k];
	  target_path[k-1] = target_path[k];
	}
	--tcount;
	--i;
      }
    }
  }
}


//#########################################################################
// INI-file handlers
//#########################################################################

//*********************************************************************
// default file colors
static const wchar_t* const attr_default_list[] =
{
  L"0x02=txt:.1st,.doc,.me,.now,.txt",
  L"0x03=src:.asm,.bas,.[ch],.[ch]pp,.pas",
  L"0x04:.bat,.cmd",
  L"0x05:.map",
  L"0x0a:.lst",
  L"0x0c=exe:.com,.exe",
  L"0x0d=inc:.hex,.inc",
  L"0x0d:.lib,.obj",
  L"0x0e=arc:.arc,.arj,.rar,.zip",
  0
};

static int write_default_ini_file( const wchar_t* ini_str )
{
  FILE* ofile;
  int	j;

  ofile = _wfopen( ini_str, L"w, ccs=UTF-8" );
  if (!ofile)
    return errno;

  fwprintf( ofile,
	    L"; Default configuration file for NDIR.\n"
	    L"; This file was generated automatically by NDIR,\n"
	    L"; but it will not be over-written.\n"
	    L"; Edit this file as you desire to re-configure NDIR.\n"
	    L";\n"
	    L"; Lines beginning with semicolons, and blank lines, are ignored.\n"
	    L"\n"
	 );

  // generate state flags
  fwprintf( ofile,
  L"ucase=%u         ; use upper-case for short names\n"
  L"sort=%u          ; 0=name, 1=ext, 2=size, 3=date/time, 4=none\n"
  L"reverse=%u       ; reverse normal sort order\n"
  L"numeric=%u       ; numeric sorting\n"
  L"clear=%u         ; clear screen before display\n"
  L"pause=%u         ; pause on full screen\n"
  L"format=%u        ; output format, 0=maximum detail to 4=name only, 5=name & age \n"
  L"minimize=%u      ; minimize header/footer\n"
  L"one_level=%u     ; directory tree only shows one level of subdirectories\n"
  L"show_all=%u      ; show system and hidden files\n"
  L"files_only=%u    ; only show files, not directories\n"
  L"dir_first=%u     ; place directories first\n"
  L"exec_only=%u     ; show executables only\n"
  L"horz=%u          ; 1=list files horizontally, 2=group by extension\n"
  L"low_ascii=%u     ; don't use graphic characters\n"
  L"color=%u         ; output in color\n"
  L"showSHRfiles=%u  ; use different color for system/hidden/readonly files\n"
  L"fdate_option=%u  ; 0=LastWriteTime, 1=LastAccessTime, 2=CreationTime\n"
  L";cols=%-*d; line length for redirected output (40 to 500)\n"
  L";rows=%-*d; rows for vertical file listing (0=maximum, -ve no divider)\n",
  n.ucase, n.sort, n.reverse, n.numeric, n.clear, n.pause, n.format, n.minimize,
  n.tree_short, n.show_all, n.files_only, n.dir_first, n.exec_only, n.horz,
  n.low_ascii, n.color, n.showSHRfiles, n.fdate_option,
  10, reportcols, 10, reportrows );

  // generate default colors
  fwprintf( ofile, L"\n"
		   L"; Global display colors.\n"
		   L"; All colors can be decimal or hex (preceded by 0x)\n"
		   L"colornhead=%u\n"
		   L"colorxhead=%u\n"
		   L"colorframe=%u\n"
		   L"colorattr=%u\n"
		   L"colorsize=%u\n"
		   L"colordate=%u\n"
		   L"colortime=%u\n"
		   L"colordir=%u\n"
		   L"colorSHR=%u\n"
		   L"colordefalt=%u\n",
		   n.colornhead, n.colorxhead, n.colorframe, n.colorattr,
		   n.colorsize, n.colordate, n.colortime, n.colordir,
		   n.colorSHR, n.colordefalt );

  // generate colors for desired extensions
  fwprintf( ofile, L"\n"
		   L"; Assign colors to file extensions and patterns.\n"
		   L"; You can add your patterns here.\n"
		   L"; Two formats are supported:\n"
		   L"; .ext=color   and\n"
		   L"; color[=id]:.ext,file.ext,.*ext*,*file*\n"
		   L"; Patterns are UTF-8, but \"id\" is only ASCII (up to four characters).\n" );
  for (j = 0; attr_default_list[j]; ++j)
  {
    fwprintf( ofile, L"%s\n", attr_default_list[j] );
    wcscpy( tempstr, attr_default_list[j] );
    parse_color_entry( tempstr );
  }

  // generate directory-tree colors
  fwprintf( ofile, L"\n"
		   L"; Assigned colors for levels in directory tree.\n"
		   L"; Each line defines an additional level, limit of 20.\n"
		   L"; Colors will wrap around based on how many are defined.\n" );
  for (j = 0; j < max_color_level; j++)
    fwprintf( ofile, L":%u\n", dtree_colors[j] );

  // generate default output formats
  fwprintf( ofile, L"\n"
		   L"; Default output formats for each column option.\n"
		   L"; Uncomment and use /#N to use a modified format.\n"
		   L"; -N is only for 8.3 names (must be first); +N for long and short.\n" );
  for (j = 0; j < 6; j++)
  {
    fwprintf( ofile, L";-%d=%s\n", j + 1, fprint[j][0] );
    fwprintf( ofile, L";+%d=%s\n", j + 1, fprint[j][1] );
  }

  fclose( ofile );
  return 0;
}

//*********************************************************************
struct ini_entry
{
  const wchar_t* lvalue;
  uchar* rvalue;
};
static const ini_entry ndir_ini[] =
{
  { L"clear",           &n.clear        },
  { L"color",           &n.color        },
  { L"colorattr",       &n.colorattr    },
  { L"colordate",       &n.colordate    },
  { L"colordefalt",     &n.colordefalt  },
  { L"colordir",        &n.colordir     },
  { L"colorframe",      &n.colorframe   },
  { L"colornhead",      &n.colornhead   },
  { L"colorSHR",        &n.colorSHR     },
  { L"colorsize",       &n.colorsize    },
  { L"colortime",       &n.colortime    },
  { L"colorxhead",      &n.colorxhead   },
  { L"dir_first",       &n.dir_first    },
  { L"exec_only",       &n.exec_only    },
  { L"fdate_option",    &n.fdate_option },
  { L"files_only",      &n.files_only   },
  { L"format",          &n.format       },
  { L"horz",            &n.horz         },
  { L"lfn_off",         &n.lfn_off      },
  { L"low_ascii",       &n.low_ascii    },
  { L"minimize",        &n.minimize     },
  { L"numeric",         &n.numeric      },
  { L"one_level",       &n.tree_short   },
  { L"pause",           &n.pause        },
  { L"reverse",         &n.reverse      },
  { L"showSHRfiles",    &n.showSHRfiles },
  { L"show_all",        &n.show_all     },
  { L"sort",            &n.sort         },
  { L"ucase",           &n.ucase        },
};

static int ini_cmp( const void* a, const void* b )
{
  return _wcsicmp( (const wchar_t*)a, ((const ini_entry*)b)->lvalue );
}

static void parse_ini_line( wchar_t* iniptr )
{
  wchar_t* eqptr;
  int	   j;

  eqptr = wcschr( iniptr, '=' );
  if (!eqptr)
    return;
  *eqptr++ = '\0';      // NULL-terminate lvalue, point to rvalue
  ini_entry* ini = (ini_entry*)bsearch( iniptr, ndir_ini,
					sizeof(ndir_ini)/sizeof(*ndir_ini),
					sizeof(*ndir_ini), ini_cmp );
  if (ini != NULL)
  {
    // if we find the lvalue (pointed to by iniptr) in the
    // lookup table, assign the rvalue (pointed to by eqptr)
    // to the variable stored
    *ini->rvalue = (uchar)wcstoul( eqptr, 0, 0 );
    // swap formats 4 & 5, to remain compatible with earlier versions
    if (ini->rvalue == &n.format)
    {
      if (n.format == 4)
	n.format = 5;
      else if (n.format == 5)
	n.format = 4;
    }
    return;
  }
  // Didn't find it, try the report columns option.
  if (_wcsicmp( iniptr, L"cols" ) == 0)
  {
    j = wcstoul( eqptr, 0, 10 );
    if (j >= 40 && j <= TEMPSTR_COUNT-2)
      reportcols = j;
  }
  // How about the report rows option?
  else if (_wcsicmp( iniptr, L"rows" ) == 0)
    reportrows = wcstol( eqptr, 0, 10 );
}

//*********************************************************************
attrib_list::attrib_list( const wchar_t* e, uchar a, DWORD i )
: next( 0 ), id( i ), attr( a )
{
  ext = _wcsdup( e );
  if (!ext)
    error_exit( OUT_OF_MEMORY, NULL );

  int idx = which_attrib_table( e );
  if (!attr_table[idx])
    attr_table[idx] = this;
  else
    attr_tail[idx]->next = this;
  attr_tail[idx] = this;
}

//*********************************************************************
int which_attrib_table( const wchar_t* e )
{
  int idx, ofs = 0;

  if (*e == '.')
  {
    ++e;
    ofs = 28;
  }
  if (is_pattern( e ))
  {
    idx = 27;
  }
  else
  {
    wchar_t c = towlower( *e );
    idx = (c >= 'a' && c <= 'z') ? c - 'a' + 1 : 0;
  }

  return idx + ofs;
}

//*********************************************************************
static void parse_color_entry( wchar_t* iniptr )
{
  wchar_t* eqptr;
  wchar_t* hdptr;
  wchar_t* tlptr;
  uchar    atr;
  DWORD    id;
  attrib_list* aptr;

  // check for multiple-color-entry forms...
  // 0x32[=id]:.com,.bat,.btm,.sys
  hdptr = wcschr( iniptr, ':' );
  if (hdptr)
  {
    atr = (uchar)wcstoul( iniptr, &eqptr, 0 );
    if (atr == 0)
      return;

    id = 0;
    if (*eqptr == '=')
    {
      while (++eqptr < hdptr)
	id = (id << 8) + (uchar)towlower( *eqptr );
    }

    while (1)
    {
      tlptr = wcschr( ++hdptr, ',' );
      // assume we're at end of line
      if (!tlptr)
      {
	new attrib_list( hdptr, atr, id );
	return;
      }
      else
      {
	*tlptr = '\0';
	aptr = new attrib_list( hdptr, atr, id );
	if (!aptr)
	  return;
      }
      hdptr = tlptr;
    }
  }
  else
  {
    eqptr = wcschr( iniptr, '=' );
    if (!eqptr)
      return;
    *eqptr++ = '\0';	// NULL-terminate lvalue, point to rvalue

    // .ARC=14
    atr = (uchar)wcstoul( eqptr, 0, 0 );
    if (atr == 0)
      return;
    new attrib_list( iniptr, atr, 0 );
  }
}

//*********************************************************************
static int dcIdx;

static void parse_dir_color_entry( wchar_t* iniptr )
{
  iniptr++;  // skip colon flag
  if (dcIdx < MAX_DIR_ENTRY)
    dtree_colors[dcIdx++] = (uchar)wcstoul( iniptr, 0, 0 );
}

//*********************************************************************
static int read_ini_file( const wchar_t* ini_str, const wchar_t* fmt )
{
  FILE*    ofile;
  int	   slen;
  wchar_t* strptr;
  wchar_t  line[MAXPATH];

  ofile = _wfopen( ini_str, L"r, ccs=UTF-8" );
  if (!ofile)
    return errno;

  dcIdx = 0;
  while (fgetws( line, sizeof(line)/sizeof(*line), ofile ))
  {
    // strip off newline char
    slen = wcslen( line );
    if (line[slen-1] == '\n')
      line[slen-1] = '\0';

    // next, find and strip off comments
    strptr = wcschr( line, ';' );
    if (strptr)
      *strptr = '\0';
    else
      strptr = line + slen;

    // then strip off trailing spaces
    while (--strptr >= line && (*strptr == ' ' || *strptr == '\t'))
      *strptr = '\0';

    // skip blank lines
    if (++strptr == line)
      continue;

    // now we should have a simple line in field=value format.
    // See if we can parse it...

    // see whether we're dealing with an extension-color entry,
    // or a flags entry
    if (fmt == NULL)
    {
      if (line[0] == '!')
      {
	insert_target_filespec( line );
      }
      else if (line[0] == '.' || (line[0] >= '0' && line[0] <= '9'))
      {
	parse_color_entry( line );
      }
      else if (line[0] == ':')
      {
	parse_dir_color_entry( line );
      }
      else if (line[0] == '#')
      {
	batchfmt = _wcsdup( line );
	n.format = 6;
      }
      else
      {
	parse_ini_line( line );
      }
    }
    else if (line[0] == '+' || (line[0] == '-' && n.lfn_off))
    {
      const wchar_t* f = fmt;
      const wchar_t* l = line + 1;
      while (towlower( *f ) == towlower( *l ) && *l != '\0')
      {
	++f;
	++l;
      }
      if ((*f == '\0' || *f == '/') && *l == '=')
      {
	batchfmt = _wcsdup( l + 1 );
	break;
      }
    }
  }
  if (dcIdx)
    max_color_level = dcIdx;

  // Explicitly require /k for redirected color output.
  if (redirected && fmt == NULL)
    n.color = 0;

  fclose( ofile );
  return 0;
}

//*********************************************************************
static void read_config_file( void )
{
  // search for existing file.
  // 1. look in current directory
  // 2. if not found, search location of executable
  // 3. if not found, generate default file in location of executable

  if (read_ini_file( local_ini_name, NULL ) == 0)
    return;

  if (read_ini_file( ininame, NULL ) == 0)
    return;

  write_default_ini_file( ininame );
}

//*********************************************************************
void find_ini_format( const wchar_t* fmt )
{
  if (read_ini_file( local_ini_name, fmt ) == 0)
    return;

  if (read_ini_file( ininame, fmt ) == 0)
    return;
}
