//*****************************************************************
//
// NSORT.CPP:	NDIR data-sorting routines
//
// Written by:	Daniel D. Miller  (the Derelict)
//		Jason Hood
//
// Last update: 01/15/01 21:20pm
//		16 Sep, 2014
//
//*****************************************************************

#include "ndata.hpp"

//***************  function prototypes  ***************
       int sort_name( const ffdata* a, const ffdata* b );
static int sort_ext(  const ffdata* a, const ffdata* b );
static int sort_size( const ffdata* a, const ffdata* b );
static int sort_time( const ffdata* a, const ffdata* b );
static int sort_dir(  const ffdata* a, const ffdata* b );
static ffdata* merge_sort( ffdata* c, int cnt );
static ffdata* merge( ffdata* a, ffdata* b );

//************************************************************
// A global function pointer is required by the
// sort routine.  This will point to a function which
// accepts two structure pointers as arguments, and
// returns:
//
// 	 > 0 if a > b
// 	== 0 if a == b
// 	 < 0 if a < b
//
//************************************************************
typedef int (*sort_fcn_t)( const ffdata* a, const ffdata* b );
static sort_fcn_t sort_fcn;
static sort_fcn_t sort_fn[SORT_NONE] =
{
  sort_name, sort_ext, sort_size, sort_time
};

//*********************************************************
int sort_name( const ffdata* a, const ffdata* b )
{
  wchar_t *an, *bn;
  int rc;

  if (n.numeric)
  {
    an = a->filename;
    bn = b->filename;
  prefix:
    // in order to sort by number, the prefix must be identical...
    while (*an == *bn && !iswdigit( *an ))
    {
      if (*an == '\0')
	return 0;
      ++an, ++bn;
    }
    // ...it must be digits that are different
    if (iswdigit( *an ) && iswdigit( *bn ))
    {
      wchar_t *ae, *be;
      unsigned long aN, bN;

      aN = wcstoul( an, &ae, 10 );
      bN = wcstoul( bn, &be, 10 );
      // ...and they must end with the same symbol
      if (*ae == *be && !iswalnum( *ae ))
      {
	if (aN > bN) return  1;
	if (aN < bN) return -1;
	// numbers are the same, are the digits?
	aN = ae - an;
	bN = be - bn;
	if (aN > bN) return  1;
	if (aN < bN) return -1;
	// see if there are other numbers
	an = ae;
	bn = be;
	goto prefix;
      }
    }
  }

  // consider the name up to the first period - this allows "name.ext" to
  // come before "name .ext"
  for (an = a->filename; *an != '.' && *an != '\0'; ++an) ;
  for (bn = b->filename; *bn != '.' && *bn != '\0'; ++bn) ;
  rc = CompareString( LOCALE_USER_DEFAULT,
		      NORM_IGNORECASE | NORM_IGNOREKANATYPE | NORM_IGNOREWIDTH,
		      a->filename, an - a->filename,
		      b->filename, bn - b->filename );
  if (rc == CSTR_EQUAL)
    rc = CompareString( LOCALE_USER_DEFAULT,
			NORM_IGNORECASE | NORM_IGNOREKANATYPE|NORM_IGNOREWIDTH,
			an, -1, bn, -1 );

  return rc - 2;
}

//*********************************************************
static int sort_ext( const ffdata* a, const ffdata* b )
{
  int c = _wcsicmp( a->ext, b->ext );
  return (c == 0) ? sort_name( a, b ) : c;
}

//*********************************************************
static int sort_size( const ffdata* a, const ffdata* b )
{
  if (a->fsize > b->fsize) return  1;
  if (a->fsize < b->fsize) return -1;
  return sort_name( a, b );
}

//*********************************************************
static int sort_time( const ffdata* a, const ffdata* b )
{
  int c = CompareFileTime( &a->ft, &b->ft );
  return (c == 0) ? sort_name( a, b ) : c;
}

//*********************************************************
// this function returns
//	< 0 if (a) is DIR and (b) is not
//	> 0 if (b) is DIR and (a) is not
//     == 0 if (a) and (b) are same type (DIR or not)
//*********************************************************
static int sort_dir( const ffdata* a, const ffdata* b )
{
  if (a->dirflag && !b->dirflag) return -1;
  if (b->dirflag && !a->dirflag) return  1;
  return 0;
}

//*********************************************************
// This routine recursively splits linked lists
// into two parts, passing the divided lists to
// merge() to merge the two sorted lists.
//*********************************************************
static ffdata* merge_sort( ffdata* a, int cnt )
{
  ffdata* b;
  static ffdata* c;
  static int n;

  if (cnt <= 1)
    return a;

  c = a;
  n = cnt >> 1;
  while (--n != 0)
    c = c->next;
  b = c->next;
  c->next = NULL;

  return merge( merge_sort( a, cnt >> 1 ),
		merge_sort( b, (cnt >> 1) + (cnt & 1) ) );
}

//*********************************************************
// This routine merges two sorted linked lists.
//*********************************************************
static ffdata* merge( ffdata* a, ffdata* b )
{
  ffdata* c;
  ffdata  z;
  int	  x;

  c = &z;
  for (;;)
  {
    x = (n.dir_first) ? sort_dir( a, b ) : 0;
    if (x == 0)
    {
      x = sort_fcn( a, b );
      if (n.reverse) x = -x;
    }
    if (x <= 0)
    {
      c->next = a;
      c = a;
      a = a->next;
      if (a == NULL)
      {
	c->next = b;
	break;
      }
    }
    else
    {
      c->next = b;
      c = b;
      b = b->next;
      if (b == NULL)
      {
	c->next = a;
	break;
      }
    }
  }

  return z.next;
}

//*********************************************************
// This intermediate function is used because I want
// merge_sort() to accept a passed parameter,
// but in this particular application the initial
// list is global.  This function sets up the global
// comparison-function pointer and passes the global
// list pointer to merge_sort().
//*********************************************************
static void sort_files( sort_fcn_t current_sort )
{
  sort_fcn = current_sort;
  ftop = merge_sort( ftop, filecount );
}

//*****************************************************************
void sort_filelist( void )
{
  if (n.sort == SORT_NONE)
  {
    if (n.dir_first) sort_files( sort_dir );
  }
  else
  {
    sort_files( sort_fn[n.sort] );

    // remove duplicated names, assuming they've been placed together
    ffdata* ftemp = ftop;
    ffdata* fnext;

    // assume at least two names
    do
    {
      // if we have a match, delete the second name
      if (wcscmp( ftemp->filename, ftemp->next->filename ) == 0)
      {
	fnext = ftemp->next->next;
	delete ftemp->next->filename;
	delete ftemp->next;
	ftemp->next = fnext;
	--filecount;
      }

      // otherwise, just move to the next file
      else
      {
	ftemp = ftemp->next;
      }
    } while (ftemp->next);
  }
}
