/******************************************************************/
/*                                                                */
/* QUALIFY.CPP: NDIR data-sorting routines			  */
/*								  */
/* Written by:	Daniel D. Miller  (the Derelict)		  */
/*		Jason Hood					  */
/*								  */
/* Last update: ?						  */
/*		17 Dec, 2005					  */
/*                                                                */
/*  qualify() modifies a string as required to generate a         */
/*  "fully-qualified" filename, which is a filename with          */
/*  drive specifier and path.					  */
/*                                                                */
/*  input:  argptr: the input filename.                           */
/*                                                                */
/*  output: qresult, a bit-mapped unsigned int with the           */
/*                   following definitions:                       */
/*                                                                */
/*          bit 0 == 1 if wildcards are present.                  */
/*          bit 1 == 1 if no wildcards and path does not exist.   */
/*          bit 2 == 1 if no wildcards and path exists as a file. */
/*          bit 7 == 1 if specified drive is invalid.             */
/*                                                                */
/******************************************************************/

#include "ndata.hpp"
#include <shlwapi.h>

/***********************  qualify() constants  ************************/
#define QUAL_WILDCARDS	0x01
#define QUAL_NO_PATH	0x02
#define QUAL_IS_FILE	0x04
#define QUAL_INV_DRIVE	0x80

/******************************************************************/
unsigned qualify( wchar_t* argptr )
{
  wchar_t  pathptr[MAXPATH+2];
  DWORD    plen;
  unsigned qresult = 0;

  if (*argptr == '\0' || (*argptr == '.' && argptr[1] == '\0'))
  {
    plen = GetCurrentDirectory( MAXPATH, argptr );
    if (argptr[plen-1] == '\\')
      --plen;
    wcscpy( argptr + plen, L"\\*" );
    return QUAL_WILDCARDS;
  }

  //******************************************************
  // get expanded path (this doesn't support UNC)
  //******************************************************
  plen = GetFullPathName( argptr, MAXPATH, pathptr, NULL );
  if (plen == 0)
    return QUAL_INV_DRIVE;

  if (plen == 3)
  {
    if (GetDriveType( pathptr ) == 1)
      return QUAL_INV_DRIVE;
    wcscpy( pathptr + 3, L"*" );
    qresult = QUAL_WILDCARDS;
  }
  else
  {
    // See if there are wildcards in argument.
    // If not, see whether path is a directory or a file,
    // or nothing.  If directory, append wildcard char.
    if (!is_pattern( pathptr ))
    {
      if (pathptr[plen-1] == '\\')
	pathptr[--plen] = '\0';

      if (PathIsDirectory( pathptr ))
      {
	wcscpy( pathptr + plen, L"\\*" );
	qresult = QUAL_WILDCARDS;	// wildcards are present
      }
      else if (PathFileExists( pathptr ) ||
	       GetLastError() == ERROR_FILE_NOT_FOUND)
      {
	qresult = QUAL_IS_FILE; 	// path exists as a normal file
      }
      else
      {
	qresult = QUAL_INV_DRIVE;	// path doesn't exist
      }
    }
    else
    {
      HANDLE handle = FindFirstFile( pathptr, &fdata );
      if (handle == INVALID_HANDLE_VALUE)
      {
	qresult = (GetLastError() == ERROR_PATH_NOT_FOUND)
		  ? QUAL_INV_DRIVE : QUAL_WILDCARDS;
      }
      else
      {
	FindClose( handle );
	qresult = QUAL_WILDCARDS;	// wildcards are present
      }
    }
  }

  //**********************************************
  // copy string back to input, then return
  //**********************************************
  wcscpy( argptr, pathptr );

  return qresult;
}
