//*****************************************************************
//
// TREELIST.CPP: Read and display directory tree
//
// Written by:	 Daniel D. Miller  (the Derelict)
//		 Jason Hood
//
// Last update:  02/07/03 18:38
//		 28 Nov, 2012
//
//*****************************************************************

#include <conio.h>
#include "ndata.hpp"
#include "conio32.hpp"

// from nsort.cpp
extern int sort_name( const ffdata* a, const ffdata* b );

static wchar_t z_root[] = L"<root>";

static const wchar_t* const dhdr[2] =
{
  L"\x251c\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500"
   L"\x252c\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500"
   L"\x253c\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500"
   L"\x252c\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500\x2500",
  L"+-----------+-----------+--------------+--------------"
};

static const wchar_t formchr[2][5] =
{
  { L'\x251c', L'\x2514', L'\x2502', L'\x2500', L'\x2524' },
  { '+', '\\', '|', '-', '+' }
};

static void sort_trees( void );

static wchar_t formstr[64*2];
static int max_display_level;

//*****************************************************************
// from FILELIST.CPP
extern void put_disk_summary( void );

//**********************************************************
// directory structure for directory_tree routines
//**********************************************************
struct dirs
{
  dirs*    brothers;
  dirs*    sons;
  wchar_t* name;
  __int64  dirsize;
  __int64  dirsecsize;
  __int64  subdirsize;
  __int64  subdirsecsize;
  int	   files;
  int	   directs;
  int	   subfiles;
  int	   subdirects;
};

//************************************************************
// A global function pointer is required by the
// sort routine.  This will point to a function which
// accepts two structure pointers as arguments, and
// return:
//
//	 > 0 if a > b
//	== 0 if a == b
//	 < 0 if a < b
//
//************************************************************
static int (*tree_sort_fcn)( const dirs* a, const dirs* b );

static dirs* top;

//**********************************************************
static wchar_t dirpath[MAXPATH];
static int     dirlen;
static int     level, colrlevel, displevel;

//***************  function prototypes  ***************
static int   build_dir_tree( void );
static int   read_dir_tree( dirs* cur_node );
static void  draw_dir_tree( void );
static void  display_dir_tree( const dirs* top );
static void  printdirheader( void );
static void  print_dir_end( void );
static dirs* new_dir_node( void );
static void  delete_dir_tree( dirs* top );

//*****************************************************************
void tree_listing( void )
{
  // release any previous dir tree
  delete_dir_tree( top );

  // read and build the dir tree
  build_dir_tree();

  // sort the tree list
  sort_trees();

  // now display the resulting directory tree
  draw_dir_tree();
}

//*********************************************************
// "waiting" pattern generator
//*********************************************************
static int dircount;

static void pattern_init( const wchar_t* lstr )
{
  dircount = 0;
  if (!hStdOut)
    fputws( lstr, stderr );
  else
    dputs( lstr );
}

static void pattern_reset( void )
{
  if (!hStdOut)
    fputws( L"\033[2K\r", stderr );
  else
  {
    dreturn();
    dclreol();
  }
}

static void pattern_update( void )
{
  wchar_t pucount[20];

  dircount++;
  _snwprintf( pucount, 20, L"%-5d\b\b\b\b\b", dircount );
  if (!hStdOut)
    fputws( pucount, stderr );
  else
    dputs( pucount );
}

//**********************************************************
static int build_dir_tree( void )
{
  int	   result;
  wchar_t* strptr;

  level = 0;

  // allocate struct for dir listing
  top = new_dir_node();

  // derive root path name
  if (base_len == 3)
  {
    top->name = z_root;
  }
  else
  {
    base_path[base_len-1] = '\0';
    for (strptr = base_path + base_len - 1; *--strptr != '\\'; ) ;
    top->name = _wcsdup( strptr + 1 );
    if (!top->name)
      error_exit( OUT_OF_MEMORY, NULL );
    base_path[base_len-1] = '\\';
    if (n.lfn_off && !n.ucase)
      _wcslwr( top->name );
  }

  wcscpy( dirpath, base_path );
  dirpath[base_len]   = '*';
  dirpath[base_len+1] = '\0';
  dirlen = base_len;

  pattern_init( L"wait; reading directory " );
  result = read_dir_tree( top );
  pattern_reset();

  return result;
}

//**********************************************************
// recursive routine to read directory tree
//**********************************************************
static int early_abort;

static int read_dir_tree( dirs* cur_node )
{
  dirs*    dtail = 0;
  wchar_t* strptr;
  int	   slen;
  __int64  clusters;

  if (((dircount % 50) == 0) && _kbhit())
  {
    // check for ESCAPE character
    if (get_scode() == 27)
      early_abort = 1;
  }
  // if early_abort flag gets set, return with "success" flag,
  // but without reading anything further...
  if (early_abort)
    return 0;

  pattern_update();

  // Insert next subtree level.
  // if level == 0, this is first call, and
  // dirpath is already complete
  if (level > 0)
  {
    // insert new path name
    slen = dirlen;
    dirlen += wcslen( cur_node->name ) + 1;
    wmemcpy( dirpath + slen, cur_node->name, dirlen - slen );
    wmemcpy( dirpath + dirlen - 1, L"\\*", 3 );
  }
  else
  {
    slen = dirlen;
  }

  if (!find_first( dirpath ))
  {
    dirlen = slen;
    DWORD err = GetLastError();
    if (err == ERROR_NO_MORE_FILES)
      err = 0;
    else
      cur_node->files = -1;
    return err;
  }

  // first, build tree list for current level
  ++level;

  // loop on find_next
  do
  {
    // we found a directory
    if (fdata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    {
      ++cur_node->directs;
      ++cur_node->subdirects;

      dirs* dtemp = new_dir_node();
      if (!cur_node->sons)
	cur_node->sons = dtemp;
      else
	dtail->brothers = dtemp;
      dtail = dtemp;

      strptr = (n.lfn_off && *fdata.cAlternateFileName)
	       ? fdata.cAlternateFileName : fdata.cFileName;
      dtail->name = _wcsdup( strptr );
      if (!dtail->name)
	error_exit( OUT_OF_MEMORY, NULL );

      // convert filename to lower case if appropriate
      if (n.lfn_off && !n.ucase)
	_wcslwr( dtail->name );
    }

    // we found a normal file
    else
    {
      // convert file size
      i64tol iconv;
      iconv.u[0] = fdata.nFileSizeLow;
      iconv.u[1] = fdata.nFileSizeHigh;
      cur_node->dirsize    += iconv.i;
      cur_node->subdirsize += iconv.i;

      wcscpy( dirpath + dirlen, fdata.cFileName );
      clusters = get_disk_size( dirpath, iconv.i,
				(fdata.dwFileAttributes
				 & (FILE_ATTRIBUTE_COMPRESSED|
				    FILE_ATTRIBUTE_SPARSE_FILE)) );
      cur_node->dirsecsize    += clusters;
      cur_node->subdirsecsize += clusters;

      ++cur_node->files;
      ++cur_node->subfiles;
    }

  } while (find_next());

  // next, build tree lists for subsequent levels (recursive)
  for (dirs* ktemp = cur_node->sons; ktemp; ktemp = ktemp->brothers)
  {
    if (read_dir_tree( ktemp ) == 0)
    {
      cur_node->subdirsize    += ktemp->subdirsize;
      cur_node->subdirsecsize += ktemp->subdirsecsize;
      cur_node->subfiles      += ktemp->subfiles;
      cur_node->subdirects    += ktemp->subdirects;
    }
  }

  // when done, strip name from path
  dirlen = slen;

  // restore the level number
  --level;
  return 0;
}

//**********************************************************
// allocate struct for dir listing
//**********************************************************
static dirs* new_dir_node( void )
{
  dirs* dtemp = new dirs;
  if (!dtemp)
    error_exit( OUT_OF_MEMORY, NULL );

  memset( (char*)dtemp, 0, sizeof(dirs) );
  if (!is_ntfs)
  {
    dtemp->dirsecsize	 = clbytes;
    dtemp->subdirsecsize = clbytes;
  }

  return dtemp;
}

//*********************************************************
// recursively delete old dir tree
//*********************************************************
static void delete_dir_tree( dirs* ktop )
{
  while (ktop)
  {
    delete_dir_tree( ktop->sons );
    dirs* kill = ktop;
    ktop = ktop->brothers;
    if (kill->name != z_root)
      delete kill->name;
    delete kill;
  }
}

//*********************************************************
static void draw_dir_tree( void )
{
  printdirheader();
  display_dir_tree( top );
  print_dir_end();
}

//**********************************************************
static int display_tree_filename( const dirs* name, const wchar_t* frmstr )
{
  int lvl, slen, frm = 1;

  slen = wcslen( name->name ) + displevel*2;

  for (lvl = 1; lvl <= displevel; ++lvl)
  {
    int clvl = (lvl == displevel) ? colrlevel : (lvl % max_color_level);
    nputc( dtree_colors[clvl], frmstr[lvl*2-2] );
    nputc( dtree_colors[clvl], frmstr[lvl*2-1] );
  }
  nputs( dtree_colors[colrlevel], name->name );

  // if directory name, etc., is too long, make separate line
  if (slen > cols - 15)
  {
    // do nothing
  }
  else if (slen > cols - 42)
  {
    nput_char( dtree_colors[colrlevel], ' ', cols - 15 - slen );
    nputc( n.colorframe, vline );
  }
  else if (slen > cols - 53)
  {
    nput_char( dtree_colors[colrlevel], ' ', cols - 42 - slen );
    nputc( n.colorframe, vline );
    nput_char( dtree_colors[colrlevel], ' ', cols - 53 );
    nputc( n.colorframe, vline );
  }
  else
  {
    frm = 0;
  }

  if (frm)
  {
    ncrlf();
    for (slen = 1; slen < displevel; ++slen)
    {
      nputc( dtree_colors[slen % max_color_level], frmstr[slen*2-2] );
      nputc( dtree_colors[slen % max_color_level], frmstr[slen*2-1] );
    }
    slen = displevel*2-1;
    nputc( dtree_colors[colrlevel],
	   (name->brothers) ? formchr[n.low_ascii][2] : ' ' );
    if (name->sons)
    {
      nputc( dtree_colors[colrlevel], ' ' );
      nputc( dtree_colors[colrlevel], formchr[n.low_ascii][2] );
      slen += 2;
    }
  }

  nput_char( dtree_colors[colrlevel], ' ', cols - 53 - slen );
  return cols - 53 - slen;
}

//**********************************************************
static void display_size( __int64 dlen, int slen, uchar attr, bool hex = true)
{
  static const wchar_t sfx[] = L" KMG";
  wchar_t* num;
  int i;
  if (hex && !n.hex)
    hex = false;
  if (hex)
  {
    num = i64toss( dlen );
    while ((int)wcslen( num ) > slen)
    {
      wchar_t* s = wcsrchr( num, ',' );
      if (s == NULL)
	break;
      wcscpy( s, s + 1 );
    }
    i = 0;
  }
  else
  {
    uchar oldhex = n.hex;
    n.hex = false;
    for (i = 0;; ++i)
    {
      num = i64toss( dlen );
      if ((int)wcslen( num ) <= slen || sfx[i] == '\0')
	break;
      if (i == 0)
	--slen;
      dlen /= 1000;
    }
    n.hex = oldhex;
  }
  nputf( attr, L"%*s", slen, num );
  if (i != 0)
    nputc( attr ^ 0x08, sfx[i] );
}

//**********************************************************
// recursive routine to display directory tree
// do all subroutines, then go to next.
//**********************************************************
static void display_dir_tree( const dirs* ktop )
{
  int size1;

  for (const dirs* ktemp = ktop; ktemp; ktemp = ktemp->brothers)
  {
    colrlevel = level % max_color_level;
    displevel = level;
    if (displevel >= max_display_level)
      displevel = max_display_level - 1;

    // first, build tree list for current level
    if (level > 0)
    {
      formstr[displevel*2-2] = formchr[n.low_ascii][(ktemp->brothers == 0)];
      formstr[displevel*2-1] = formchr[n.low_ascii][3];
    }
    formstr[displevel*2] = '\0';

    //*****************************************************************
    // display data for this level
    //*****************************************************************
    size1 = display_tree_filename( ktemp, formstr );
    if (size1 > 0)
      size1 = 0;
    if (ktemp->files == -1)
    {
      nput_char( dtree_colors[colrlevel], ' ', 11 + size1 );
      nputc( n.colorframe, vline );
      nput_char( dtree_colors[colrlevel], ' ', 26 );
      nputc( n.colorframe, vline );
      nput_char( dtree_colors[colrlevel], ' ', (n.tree == 2) ? 4 : 8 );
      nputs( dtree_colors[colrlevel], L"Error!" );
    }
    else switch (n.tree)
    {
      // show file/directory sizes only
      case 1:
	if (ktemp->dirsize    != ktemp->subdirsize ||
	    ktemp->dirsecsize != ktemp->subdirsecsize)
	{
	  display_size( ktemp->dirsize, 11 + size1, dtree_colors[colrlevel] );
	  nputc( n.colorframe, vline );
	  display_size( ktemp->dirsecsize, 11, dtree_colors[colrlevel] );
	  nputc( dtree_colors[colrlevel], ' ' );
	}
	// no subdirectories are under this one
	else
	{
	  nput_char( dtree_colors[colrlevel], ' ', 11 + size1 );
	  nputc( n.colorframe, vline );
	  nput_char( dtree_colors[colrlevel], ' ', 12 );
	}
	display_size( ktemp->subdirsize, 14, dtree_colors[colrlevel] );
	nputc( n.colorframe, vline );
	display_size( ktemp->subdirsecsize, 14, dtree_colors[colrlevel] );
      break;

      // show file/directory counts only
      case 2:
	if (ktemp->files   != ktemp->subfiles ||
	    ktemp->directs != ktemp->subdirects)
	{
	  display_size( (__int64)ktemp->files, 9 + size1,
			dtree_colors[colrlevel], false );
	  nputs( dtree_colors[colrlevel], L"  " );
	  nputc( n.colorframe, vline );
	  display_size( (__int64)ktemp->directs, 9, dtree_colors[colrlevel],
			false );
	  nputs( dtree_colors[colrlevel], L"   " );
	}
	// no subdirectories are under this one
	else
	{
	  nput_char( dtree_colors[colrlevel], ' ', 11 + size1 );
	  nputc( n.colorframe, vline );
	  nput_char( dtree_colors[colrlevel], ' ', 12 );
	}
	display_size( (__int64)ktemp->subfiles, 12, dtree_colors[colrlevel],
		      false );
	nputs( dtree_colors[colrlevel], L"  " );
	nputc( n.colorframe, vline );
	display_size( (__int64)ktemp->subdirects, 10, dtree_colors[colrlevel],
		      false );
      break;

      // show mixed size, file counts, directory counts
      case 3:
	display_size( (__int64)ktemp->subfiles, 9 + size1,
		      dtree_colors[colrlevel], false );
	nputs( dtree_colors[colrlevel], L"  " );
	nputc( n.colorframe, vline );
	display_size( (__int64)ktemp->subdirects, 9, dtree_colors[colrlevel],
		      false );
	nputs( dtree_colors[colrlevel], L"   " );
	display_size( ktemp->dirsecsize, 14, dtree_colors[colrlevel] );
	nputc( n.colorframe, vline );
	display_size( ktemp->subdirsecsize, 14, dtree_colors[colrlevel] );
      break;
    }
    ncrlf();
    ++filesdisp;

    // build tree string for deeper levels
    if (level > 0)
    {
      formstr[displevel*2-2] = (ktemp->brothers) ? formchr[n.low_ascii][2] : ' ';
      formstr[displevel*2-1] = ' ';
    }

    // process any sons
    ++level;
    if (!n.tree_short || level < 2)
      display_dir_tree( ktemp->sons );

    if (--level < max_display_level)
      formstr[level*2] = '\0';
  } // while not done listing directories
}

/*****************************************************************/
static void printdirheader( void )
{
  int slen;

  if (cols < 79)
    cols = 79;
  max_display_level = (cols - 42 - 4) / 2 + 1;

  nputs( n.colornhead, L"Directory Tree of " );
  nputs( n.colorxhead, base_path );

  slen = 18 + base_len + 1 + 16 + wcslen( volume_name );
  if (slen <= cols)
  {
    nput_char( n.colornhead, ' ', cols - slen + 1 );
    nputs( n.colornhead, L"Volume label is " );
    nputs( n.colorxhead, volume_name );
  }
  ncrlf();

  //**************************************
  // Heading line 1
  //**************************************
  wmemset( tempstr, dline, cols );
  tempstr[cols] = '\0';
  tempstr[cols-54] = tempstr[cols-30] = tline;
  nputs( n.colorframe, tempstr );
  ncrlf();

  const wchar_t* hd[4];
  switch (n.tree)
  {
    default:
    case 1:
      hd[0] = L"   size of requested   ";
      hd[2] = L"       directory       ";
      hd[1] = L"    total size,  including";
      hd[3] = L"     lower subdirectories";
    break;

    case 2:
      hd[0] = L" files and directories ";
      hd[2] = L" in current directory  ";
      hd[1] = L"    files and directories";
      hd[3] = L"  in current and lower dirs";
    break;

    case 3:
      hd[0] = L" files and directories ";
      hd[2] = L"   cumulative counts   ";
      hd[1] = L"    files and directories";
      hd[3] = L"       cumulative sizes";
    break;
  }
  //**************************************
  // Heading line 2
  //**************************************
  nput_char( n.colornhead, ' ', cols - 54 );
  nputc( n.colorframe, vline );
  nputs( n.colornhead, hd[0] );
  nputc( n.colorframe, vline );
  nputs( n.colornhead, hd[1] );
  ncrlf();

  //**************************************
  // Heading line 3
  //**************************************
  nput_char( n.colornhead, ' ', cols - 54 );
  nputc( n.colorframe, vline );
  nputs( n.colornhead, hd[2] );
  nputc( n.colorframe, vline );
  nputs( n.colornhead, hd[3] );
  ncrlf();

  //**************************************
  // Heading line 4
  //**************************************
  nput_char( n.colorframe, ' ', cols - 54 );
  nputs( n.colorframe, dhdr[n.low_ascii] );
  ncrlf();

  //**************************************
  // Heading line 5
  //**************************************
  nputs( n.colornhead, L"Subdirectory names" );
  nput_char( n.colornhead, ' ', cols - 54 - 18 );
  nputc( n.colorframe, vline );
  switch (n.tree)
  {
    default:
    case 1:
      hd[0] = L"  in bytes ";
      hd[1] = L" disk space";
      hd[2] = L"   in bytes   ";
      hd[3] = L"  disk space";
    break;

    case 2:
      hd[0] = L"    files  ";
      hd[1] = L"directories";
      hd[2] = L"       files  ";
      hd[3] = L" directories";
    break;

    case 3:
      hd[0] = L"    files  ";
      hd[1] = L"directories";
      hd[2] = L"  disk space  ";
      hd[3] = L"  disk space";
    break;
  }
  nputs( n.colornhead, hd[0] );
  nputc( n.colorframe, vline );
  nputs( n.colornhead, hd[1] );
  nputc( n.colorframe, vline );
  nputs( n.colornhead, hd[2] );
  nputc( n.colorframe, vline );
  nputs( n.colornhead, hd[3] );
  ncrlf();

  //**************************************
  // Heading line 6
  //**************************************
  tempstr[cols-54] = tempstr[cols-30] = bline;
  tempstr[cols-42] = tempstr[cols-15] = xline;
  nputs( n.colorframe, tempstr );
  ncrlf();

  filesdisp = 1;
  filecount = ((n.tree_short) ? top->directs : top->subdirects) + 1;
}

//*********************************************************
static void print_dir_end( void )
{
  // draw divider line for bottom of data
  wmemset( tempstr, dline, cols );
  tempstr[cols] = '\0';
  tempstr[cols-42] = tempstr[cols-15] = bline;
  nputs( n.colorframe, tempstr );
  ncrlf();

  // now show disk totals
  put_disk_summary();
}

//*********************************************************
static int tree_sort_name( const dirs* a, const dirs* b )
{
  ffdata af, bf;
  af.filename = a->name;
  bf.filename = b->name;
  return sort_name( &af, &bf );
}

//*********************************************************
static int tree_sort_size( const dirs* a, const dirs* b )
{
  if (a->subdirsecsize > b->subdirsecsize) return  1;
  if (a->subdirsecsize < b->subdirsecsize) return -1;
  return 0;
}

//*********************************************************
// This routine merges two sorted linked lists.
//*********************************************************
static dirs* tree_merge( dirs* a, dirs* b )
{
  dirs* c;
  dirs	z;

  c = &z;
  for (;;)
  {
    int x = tree_sort_fcn( a, b );
    if (n.reverse) x = -x;
    if (x <= 0)
    {
      c->brothers = a;
      c = a;
      a = a->brothers;
      if (!a)
      {
	c->brothers = b;
	break;
      }
    }
    else
    {
      c->brothers = b;
      c = b;
      b = b->brothers;
      if (!b)
      {
	c->brothers = a;
	break;
      }
    }
  }

  return z.brothers;
}

//*********************************************************
// This routine recursively splits linked lists
// into two parts, passing the divided lists to
// merge() to merge the two sorted lists.
//*********************************************************
static dirs* tree_merge_sort( dirs* a, int directs )
{
  dirs* b;
  static dirs* c;
  static int   j = 0;

  if (directs < 2)
    return a;

  c = a;
  j = directs >> 1;
  while (--j != 0)
    c = c->brothers;
  b = c->brothers;
  c->brothers = 0;

  return tree_merge( tree_merge_sort( a, directs >> 1 ),
		     tree_merge_sort( b, (directs >> 1) + (directs & 1) ) );
}

//*********************************************************
// recursive routine that sorts the sons of each brother
//*********************************************************
static dirs* tree_sort_walk( dirs* t, unsigned directs )
{
  for (dirs* dptr = t; dptr; dptr = dptr->brothers)
    dptr->sons = tree_sort_walk( dptr->sons, dptr->directs );

  return tree_merge_sort( t, directs );
}

//*********************************************************
// This intermediate function is used because I want
// merge_sort() to accept a passed parameter,
// but in this particular application the initial
// list is global.  This function sets up the global
// comparison-function pointer and passes the global
// list pointer to merge_sort().
//*********************************************************
static void sort_trees( void )
{
  tree_sort_fcn = (n.sort == SORT_SIZE) ? tree_sort_size : tree_sort_name;

  // now, sort the data
  top = tree_sort_walk( top, 0 );
}
